# -*- coding: utf-8 -*-

import os
import sys
import tempfile
import tarfile
import zipfile
from boto.compat import json
import subprocess
import boto
import boto.emr.emrobject
import boto.emr.step
from urlparse import urlparse, urlsplit, urlunsplit, urlunparse, unquote
import urllib
from json import loads as json_loads
import copy


def clean_hdfs_path(p):
    if cmd("hadoop fs -rm -r -f %s && hadoop fs -mkdir -p %s" % (p, p)) == 0:
        return True
    else:
        return False


def hdfs_safe_upload(f_local, f_remote):
    f_remote_dir = os.path.dirname(f_remote)
    if cmd("hadoop fs -mkdir -p %s" % f_remote_dir) != 0:
        raise Exception("Failed to create dir %s" % f_remote_dir)
    print("HDFS Upload :: %s ====> %s" % (f_local, f_remote))
    if cmd("hadoop fs -copyFromLocal %s %s" % (f_local, f_remote_dir)) != 0:
        raise Exception("Failed to upload file %s to %s" % (f_local, f_remote_dir))


def cmd(cmd_str):
    print("Execute External Command : '%s'" % cmd_str)
    ret = subprocess.call(cmd_str, shell=True)
    print("Exit with exit code = %d" % ret)
    return ret


def s3_upload(bucket, local_filename, remote_filename):

    if urlparse(local_filename).scheme in ["s3", "s3n"]:
        return local_filename

    # max size in bytes before uploading in parts.
    # between 1 and 5 GB recommended
    MAX_SIZE = 400 * 1000 * 1000
    # size of parts when uploading in parts
    PART_SIZE = 6 * 1000 * 1000

    fn_local = os.path.normpath(local_filename)
    fn_remote = urlparse(remote_filename).path
    fn_remote_full = remote_filename

    filesize = os.path.getsize(local_filename)
    print("filesize = %d, maxsize = %d" % (filesize, MAX_SIZE))
    if filesize > MAX_SIZE:
        print("Multi-part uploading...")
        print("From : %s" % fn_local)
        print("To   : %s" % fn_remote_full)
        mp = bucket.initiate_multipart_upload(fn_local)
        fp = open(fn_local, 'rb')
        fp_num = 0
        while fp.tell() < filesize:
            fp_num += 1
            print "uploading part %i" % fp_num
            mp.upload_part_from_file(fp, fp_num, cb=percent_cb, num_cb=10, size=PART_SIZE)
        mp.complete_upload()
        print("Done")
        print("")
    else:
        print("Single-part upload...")
        print("From : %s" % fn_local)
        print("To   : %s" % fn_remote_full)
        k = boto.s3.key.Key(bucket)
        k.key = fn_remote
        k.set_contents_from_filename(fn_local, cb=percent_cb, num_cb=10)
        print("")

    return fn_remote_full


def s3_delete(bucket, s3_path):
    from urlparse import urlparse

    print("s3_delete %s" % s3_path)
    prefix_path = urlparse(s3_path).path[1:]
    for key in bucket.list(prefix=prefix_path):
        key.delete()
    return True


def percent_cb(complete, total):
    sys.stdout.write('.')
    sys.stdout.flush()


_missing = None


class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """

    # implementation detail: this property is implemented as non-data
    # descriptor.  non-data descriptors are only invoked if there is
    # no entry with the same name in the instance's __dict__.
    # this allows us to completely get rid of the access function call
    # overhead.  If one choses to invoke __get__ by hand the property
    # will still work as expected because the lookup logic is replicated
    # in __get__ for manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def url_path_join(*parts):
    """Normalize url parts and join them with a slash."""
    schemes, netlocs, paths, queries, fragments = zip(*(urlsplit(part) for part in parts))
    scheme, netloc, query, fragment = first_of_each(schemes, netlocs, queries, fragments)
    path = '/'.join(x.strip('/') for x in paths if x)
    return urlunsplit((scheme, netloc, path, query, fragment))

def first_of_each(*sequences):
    return (next((x for x in sequence if x), '') for sequence in sequences)


def s3parse(s3_path):
    pr = urlparse(s3_path)
    if not pr.path:
        pr = pr._replace(path="/")
    return pr


def s3join(s3_path, *rel_path):
    pr = urlparse(s3_path)
    pr = pr._replace(path=os.path.normpath(os.path.join(pr.path, *rel_path)))
    return urlunparse(pr)

# TODO: Refactor to 'botocore' when it becomes mature.
def convert_emr_object(obj):
    """Convert a boto emrobject into human readable python dict.

    :param obj: A emrobject
    :return: dict
    """

    if not isinstance(obj, boto.emr.emrobject.EmrObject):
        return obj

    if isinstance(obj, boto.emr.emrobject.Step):
        ret_obj = obj.__dict__
        ret_obj['args'] = [convert_emr_object(arg) for arg in ret_obj['args']]
        del ret_obj['connection']
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.HadoopStep):
        ret_obj = obj.__dict__
        ret_obj['status'] = convert_emr_object(ret_obj['status'])
        ret_obj['config'] = convert_emr_object(ret_obj['config'])
        del ret_obj['connection']
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.ClusterStatus):
        ret_obj = obj.__dict__
        ret_obj['timeline'] = convert_emr_object(ret_obj['timeline'])
        ret_obj['statechangereason'] = convert_emr_object(ret_obj['statechangereason'])
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.ClusterTimeline):
        return obj.__dict__
    elif isinstance(obj, boto.emr.emrobject.StepConfig):
        ret_obj = obj.__dict__
        ret_obj['args'] = [convert_emr_object(arg) for arg in ret_obj['args']]
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.Arg):
        return obj.value
    elif isinstance(obj, boto.emr.emrobject.JobFlow):
        ret_obj = obj.__dict__
        ret_obj['bootstrapactions'] = [convert_emr_object(b) for b in ret_obj['bootstrapactions']]
        ret_obj['instancegroups'] = [convert_emr_object(b) for b in ret_obj['instancegroups']]
        ret_obj['steps'] = [convert_emr_object(b) for b in ret_obj['steps']]
        del ret_obj['connection']
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.BootstrapAction):
        ret_obj = obj.__dict__
        ret_obj['args'] = [convert_emr_object(arg) for arg in ret_obj['args']]
        del ret_obj['connection']
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.InstanceGroup):
        ret_obj = obj.__dict__
        del ret_obj['connection']
        return ret_obj
    elif isinstance(obj, boto.emr.emrobject.ClusterStateChangeReason):
        ret_obj = obj.__dict__
        return ret_obj
    else:
        return str(obj)

    # elif isinstance(obj, boto.emr.emrobject.EmrConnection):
    #     print "EmrConnection"
    #     print dir(obj)
    #     return obj

    return obj


def pprint_aws_obj(obj):
    return json.dumps(convert_emr_object(obj), sort_keys=True, indent=4, separators=(',', ': '))
    # return json.dumps(obj, cls=BotoJsonEncoder)


def pprint_json(obj):
    return json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '))


def prepare_hadoop_conf_tarfile(fn_tarfile):
    tf = tarfile.open(fn_tarfile, "r:*")
    ti_dirs = [ti.path for ti in tf if ti.isdir()]
    if len(ti_dirs) > 1:
        raise Exception("Invalid hadoop_conf archive format. It must contain zero or one dir on top level!")

    tmp_hadoop_dir = tempfile.mkdtemp(prefix="hadoop_conf_")
    for member in tf.getmembers():
        if member.isreg():
            tf.extract(member, tmp_hadoop_dir)

    final_hadoop_conf_dir = os.path.join(tmp_hadoop_dir, *ti_dirs)
    return final_hadoop_conf_dir


def zip_isdir(fn):
    return fn.endswith("/")


def prepare_hadoop_conf_zipfile(fn_zipfile):
    zf = zipfile.ZipFile(fn_zipfile)
    zi_dirs = [zi for zi in zf.namelist() if zip_isdir(zi)]
    if len(zi_dirs) > 1:
        raise Exception("Invalid hadoop_conf archive format. It must contain zero or one dir on top level!")

    tmp_hadoop_dir = tempfile.mkdtemp(prefix="hadoop_conf_")

    for zi in zf.infolist():
        zf.extract(zi, tmp_hadoop_dir)

    final_hadoop_conf_dir = os.path.join(tmp_hadoop_dir, *zi_dirs)
    return final_hadoop_conf_dir


def prepare_hadoop_conf(fn, safe=False):
    def _prepare_hadoop_conf_file(fn):

        if tarfile.is_tarfile(fn):
            return prepare_hadoop_conf_tarfile(fn)
        elif zipfile.is_zipfile(fn):
            return prepare_hadoop_conf_zipfile(fn)
        else:
            raise Exception("Can not prepare hadoop_conf archive file: '%s'" % fn)

    def _prepare_hadoop_conf(fn):
        pr = urlparse(fn)
        if pr.scheme in ["http", "https"]:
            with tempfile.NamedTemporaryFile(prefix="tmp_hadoop_conf_") as f:
                try:
                    urllib.urlretrieve(fn, f.name)
                except Exception as e:
                    raise Exception("ERROR : failed to download '%s'" % fn)
                return _prepare_hadoop_conf_file(f.name)
        elif pr.scheme in ["file"]:
            return _prepare_hadoop_conf_file(pr.path)
        elif pr.scheme == '':
            return _prepare_hadoop_conf_file(fn)
        else:
            raise Exception("Do not support file type '%s'" % fn)

    if safe:
        try:
            return _prepare_hadoop_conf(fn)
        except Exception as e:
            print "WARNING: got exception : %s" % e
            return None
    else:
        return _prepare_hadoop_conf(fn)


def preprocess_cluster_envs(base_envs, hadoop_type, cluster_def):
    base_envs = copy.copy(dict(base_envs))

    def _get_new_env(env_name, env_val):

        if isinstance(env_val, dict):
            if 'prepend_path' in env_val:
                ppath = env_val['prepend_path']
                if env_name in base_envs:
                    return ppath + os.pathsep + base_envs.get(env_name)
                else:
                    return ppath
            elif 'append_path' in env_val:
                ppath = env_val['append_path']
                if env_name in base_envs:
                    return base_envs.get(env_name) + os.pathsep + ppath
                else:
                    return ppath
            else:
                # TODO:
                pass
        elif isinstance(env_val, basestring):
            return env_val
        else:
            print "WARNING: Invalid basic env_vars from cluster_def. Only string or dict."
            return None

    if not os.path.isfile(cluster_def):
        print "WARNING: Can not find definition for clusters, use default one."
        return base_envs
    cluster_defs = json_loads(open(cluster_def).read())
    if hadoop_type not in cluster_defs:
        print "WARNING: Can not find definition of cluster '%s', use default one." % hadoop_type
        return base_envs

    for ek, ev in cluster_defs[hadoop_type]["env_vars"].items():
        ev_new = _get_new_env(ek, ev)
        if ev_new:
            # print "%s=%s" % (ek, ev_new)
            base_envs[ek] = ev_new
    return base_envs

def mask_key(s, show=5):
    return s[:show] + '*' * (len(s) - show*2) + s[-show:]



# -*- coding: utf-8 -*-

"""
A series of functions to parse hadoop logs, for example, the hadoop counter.
"""

import re
import string
import pyparsing as pyp
from collections import OrderedDict

try:
    from cStringIO import StringIO
except ImportError:  # python 3
    from io import StringIO


class SemanticGroup(object):
    """The (abstract) class to represent a semantic group of pyparsing."""

    label = "SemanticGroup"

    def __init__(self, contents):
        self.contents = contents
        while self.contents[-1].__class__ == self.__class__:
            self.contents = self.contents[:-1] + self.contents[-1].contents

    def __str__(self):
        return "%s(%s)" % (self.label,
                           " ".join([isinstance(c, basestring) and c or str(c) for c in self.contents]))

    def __repr__(self):
        return "%s(%s)" % (self.label,
                           " ".join([isinstance(c, basestring) and c or str(c) for c in self.contents]))


class HadoopCounter(SemanticGroup):
    """The class to represent a hadoop counter."""

    label = "HadoopCounter"

    def __init__(self, contents):
        super(HadoopCounter, self).__init__(contents)
        self.name = contents[0]
        self.val = contents[1]

    def __str__(self):
        return "<HadoopCounter :: '%s' = %s>" % (self.name, self.val)

    def __repr__(self):
        return "<HadoopCounter :: '%s' = %s>" % (self.name, self.val)


class HadoopCounterGroup(SemanticGroup):
    """The class to represent a hadoop counter group."""

    label = "HadoopCounterGroup"

    def __init__(self, contents):
        super(HadoopCounterGroup, self).__init__(contents)
        self.name = contents[0]
        self.group = contents[1:]

    def __str__(self):
        return "<HadoopCounterGroup :: '%s' => '%s'>" % (self.name, self.group)

    def __repr__(self):
        return "<HadoopCounterGroup :: '%s' => '%s'>" % (self.name, self.group)

    @property
    def asDict(self):
        """Return hadoop counter group as dict.

        :rtype : dict
        :return: group as dict.
        """
        return OrderedDict([(g.name, g.val) for g in self.group])


class HadoopCounterGroups(SemanticGroup):
    """The class to represent a hadoop counter groups."""

    label = "HadoopCounterGroup"

    def __init__(self, contents):
        super(HadoopCounterGroups, self).__init__(contents)
        self.groups = contents

    def __str__(self):
        return "<HadoopCounterGroups :: '%s'>" % self.groups

    def __repr__(self):
        return "<HadoopCounterGroups :: '%s'>" % self.groups

    @property
    def asDict(self):
        """Return hadoop counter groups as dict.

        :rtype : dict
        :return: groups as dict.
        """
        return OrderedDict([(g.name, g) for g in self.groups])


class SyslogMessage(SemanticGroup):
    """The class to represent a hadoop counter group."""

    label = "SyslogMessage"

    def __init__(self, contents):
        super(SyslogMessage, self).__init__(contents)
        self.msg_date = contents[0]
        self.level = contents[1]
        self.jar_pkg = contents[2]
        self.jar_main = contents[3]
        self.msg = contents[4]

    def __str__(self):
        return "<SyslogMessage :: '%s' '%s' '%s' '%s'>" % (self.msg_date, self.level, self.jar_pkg, self.jar_main)

    def __repr__(self):
        return "<SyslogMessage :: '%s' '%s' '%s' '%s'>" % (self.msg_date, self.level, self.jar_pkg, self.jar_main)


def makeGroupObject(cls):
    def groupAction(s, l, t):
        try:
            return cls(t[0].asList())
        except:
            return cls(t)

    return groupAction


# ##############################################
# Parse 'stderr' text from s3
###############################################


def syntax_hadoop_counter():
    """Build a grammer to parse hadoop counter from logs.

    :return: A pyparsing parser.
    """

    integer = pyp.Word(pyp.nums)
    words = pyp.Word(pyp.alphas + ".")
    words_ = pyp.Word(pyp.printables + " ", excludeChars="=")
    # header = pyp.Combine(words + ": " + words + ": " + integer)
    header = pyp.Literal("Counters:") + integer
    group_name = words_

    item_key = words_ + pyp.Suppress("=")
    item_value = integer
    counter_kv = pyp.Group(item_key + item_value).setParseAction(makeGroupObject(HadoopCounter))
    counter_group = pyp.Group(group_name + pyp.OneOrMore(counter_kv)) \
        .setParseAction(makeGroupObject(HadoopCounterGroup))
    counter_groups = pyp.Group(pyp.OneOrMore(counter_group))

    # bnf = header("Header") + counter_groups("Groups")
    bnf = header("Header") + counter_groups("Groups").setParseAction(makeGroupObject(HadoopCounterGroups))

    return bnf


SYSLOG_PATTERN = r"^\d{2}/\d{1,2}/\d{2,4} \d{2}:\d{2}:\d{2} [A-Z]* "


def extract_hadoop_counter(txt_block):
    """Extract hadoop counter from string of text block.

    :param txt_block: The string block to be processed.
    :return: A Hadoop group counter object.
    """
    hadoop_counter_parser = syntax_hadoop_counter()
    # r = hadoop_counter_parser.parseString(txt_block)
    r = hadoop_counter_parser.searchString(txt_block)
    return r


def parse_s3_stderr(content_text):
    """The function to parse hadoop log and extract hadoop counters.

    The format looks like this:

        ...
        14/10/22 08:08:22 INFO mapreduce.Job: Counters: 13
            Job Counters
                Failed map tasks=11
                Killed map tasks=2
                Launched map tasks=13
                Other local map tasks=9
                Rack-local map tasks=4
                Total time spent by all maps in occupied slots (ms)=871299
                Total time spent by all reduces in occupied slots (ms)=0
                Total time spent by all map tasks (ms)=290433
                Total vcore-seconds taken by all map tasks=290433
                Total megabyte-seconds taken by all map tasks=223052544
            Map-Reduce Framework
                CPU time spent (ms)=0
                Physical memory (bytes) snapshot=0
                Virtual memory (bytes) snapshot=0
        ...

    :param content_text: the string to be parsed
    :return: A Hadoop group counter object.
    """
    for l in re.split(SYSLOG_PATTERN, content_text, flags=re.MULTILINE):
        if re.match("^mapreduce.Job: Counters: \d+", l):
            return extract_hadoop_counter(l)


###############################################
# Parse syslog format
###############################################

def syntax_unified_log():
    msg_date = pyp.Regex("^\d{2,4}[-/]\d{1,2}[-/]\d{2,4} \d{2}:\d{2}:\d{2}(,\d+)?")
    level = pyp.Word(string.ascii_uppercase)
    jar_pkg = pyp.Regex("[a-zA-Z\.]+")
    jar_main = pyp.Suppress("(") + pyp.Word(pyp.printables, excludeChars="()") + pyp.Suppress(")")

    header = msg_date + level + jar_pkg + jar_main + pyp.Suppress(":")
    msg = pyp.restOfLine
    bnf = pyp.Group(header + msg).setParseAction(makeGroupObject(SyslogMessage))
    return bnf


def syslog_generator(texts, filter_condition=lambda x: True):
    p = syntax_unified_log()
    for l in StringIO(texts):
        obs = p.parseString(l)
        if len(obs) == 1:
            o = obs[0]
            if filter_condition(o):
                yield o


def parse_syslog(text):
    """This function will parse text like this:

        ...
        2014-11-19 16:04:54,677 INFO org.apache.hadoop.mapred.JobClient (main): Job complete: job_201411191552_0001
        2014-11-19 16:04:54,711 INFO org.apache.hadoop.mapred.JobClient (main): Counters: 40
        2014-11-19 16:04:54,711 INFO org.apache.hadoop.mapred.JobClient (main):   counter_group_name
        2014-11-19 16:04:54,711 INFO org.apache.hadoop.mapred.JobClient (main):     counter name 1=10
        2014-11-19 16:04:54,712 INFO org.apache.hadoop.mapred.JobClient (main):     counter name 2=10
        2014-11-19 16:04:54,712 INFO org.apache.hadoop.mapred.JobClient (main):   another_counter_group_name
        ...


    :param text: The string content to be parsed.
    :return: A Hadoop group counter object.
    """
    from itertools import dropwhile

    info_texts = syslog_generator(text, lambda x: x.level in ["INFO"])
    counter_text_block = [l.msg for l in dropwhile(lambda x: not re.search("Job complete: \w+", x.msg), info_texts)]
    return extract_hadoop_counter("\n".join(counter_text_block))



# -*- coding: utf-8 -*-

"""
A series of Runtime.
"""

import boto
import time
import itertools
from datacanvas import EmrCluster
from datacanvas.utils import *
from datacanvas.module import get_settings_from_file


class DatacanvasRuntime(object):
    def __init__(self, spec_filename="spec.json"):
        self.settings = get_settings_from_file(spec_filename)

    def __repr__(self):
        return str(self.settings)


class HadoopRuntime(DatacanvasRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(HadoopRuntime, self).__init__(spec_filename=spec_filename)

    @property
    def hdfs_root(self):
        ps = self.settings
        if 'hdfs_root' in ps.Param._asdict():
            return ps.Param.hdfs_root.val
        else:
            return '/'

    def get_hdfs_working_dir(self, path=""):
        ps = self.settings
        glb_vars = ps.GlobalParam
        remote_path = os.path.normpath(
            os.path.join('tmp/zetjob', glb_vars['userName'], "job%s" % glb_vars['jobId'], "blk%s" % glb_vars['blockId'],
                         path))
        return os.path.join(self.hdfs_root, remote_path)

    def get_hive_namespace(self):
        ps = self.settings
        glb_vars = ps.GlobalParam
        return "zetjobns_%s_job%s_blk%s" % (glb_vars['userName'], glb_vars['jobId'], glb_vars['blockId'])

    def hdfs_upload_dir(self, local_dir):
        for root_dir, dirs, files in os.walk(local_dir):
            for f in sorted(files):
                f = os.path.normpath(os.path.join(root_dir, f))
                f_remote = self.get_hdfs_working_dir(f)
                hdfs_safe_upload(f, f_remote)
                yield f_remote

    def hdfs_clean_working_dir(self):
        hdfs_working_dir = self.get_hdfs_working_dir()
        if not clean_hdfs_path(hdfs_working_dir):
            # TODO : refactor to 'HiveException'
            raise Exception("Can not clean hdfs path : %s" % hdfs_working_dir)

    def clean_working_dir(self):
        self.hdfs_clean_working_dir()


class EmrRuntime(HadoopRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrRuntime, self).__init__(spec_filename)
        p = self.settings.Param
        self.s3_conn = boto.connect_s3(p.AWS_ACCESS_KEY_ID, p.AWS_ACCESS_KEY_SECRET)
        self.s3_bucket = self.s3_conn.get_bucket(p.S3_BUCKET)
        self.cluster = EmrCluster(p.AWS_Region, p.AWS_ACCESS_KEY_ID, p.AWS_ACCESS_KEY_SECRET, p.EMR_jobFlowId)

    def get_s3_working_dir(self, path=""):
        ps = self.settings
        glb_vars = ps.GlobalParam
        remote_path = os.path.normpath(
            os.path.join(self.s3_bucket.name, 'zetjob', glb_vars['userName'], "job%s" % glb_vars['jobId'],
                         "blk%s" % glb_vars['blockId'], path))
        return os.path.join("s3://", remote_path)

    def get_emr_job_name(self):
        ps = self.settings
        glb_vars = ps.GlobalParam
        return os.path.join('zetjob', glb_vars['userName'], "job%s" % glb_vars['jobId'], "blk%s" % glb_vars['blockId'])

    def s3_upload_dir(self, local_dir):
        print("EmrHiveRuntime.s3_uploader()")
        print("s3_upload_dir :::: %s" % local_dir)

        if local_dir == "":
            return

        if not os.path.isdir(local_dir):
            return

        s3_upload_dir = self.get_s3_working_dir(local_dir)
        ext_files = [f for f in sorted(os.listdir(local_dir)) if os.path.isfile(os.path.join(local_dir, f))]
        for f in ext_files:
            f_local = os.path.join(local_dir, f)
            f_remote_full = self.get_s3_working_dir(os.path.join(local_dir, f))

            print("S3 Upload      :: %s ====> %s" % (f_local, s3_upload_dir))
            print("S3 remote_full :: %s" % f_remote_full)
            yield s3_upload(self.s3_bucket, f_local, f_remote_full)

    def s3_clean_working_dir(self):
        s3_working_dir = self.get_s3_working_dir()
        if not s3_delete(self.s3_bucket, s3_working_dir):
            # TODO : refactor to 'HiveException'
            raise Exception("Can not clean s3 path : %s" % s3_working_dir)

    def s3_upload(self, filename):
        from urlparse import urlparse

        parse_ret = urlparse(filename)
        if parse_ret.scheme == '':
            s3_working_dir = self.get_s3_working_dir()
            file_remote = os.path.join(s3_working_dir, os.path.normpath(os.path.basename(filename)))
            file_remote_full = s3_upload(self.s3_bucket, filename, file_remote)
            return file_remote_full
        elif parse_ret.scheme == 's3':
            return filename
        else:
            raise ValueError("Invalid filename to upload to s3: %s" % filename)

    def clean_working_dir(self):
        self.s3_clean_working_dir()

    def dump_logs(self, step_id, log_files=None, retry_count=1, retry_interval=60, extra_wait_count=5):
        if not log_files:
            return

        while retry_count > 0:
            remote_log_files = [os.path.basename(i)
                                for i in self.cluster.s3_list_files(self.cluster.emr_step_log_filename(step_id))]
            intersection_files = set(log_files) & set(remote_log_files)
            if not intersection_files:
                print "Seems logs are not ready on s3, retry..."
                retry_count -= 1
                time.sleep(retry_interval)
                continue
            for i in range(extra_wait_count):
                remote_log_files = [os.path.basename(i)
                                    for i in self.cluster.s3_list_files(self.cluster.emr_step_log_filename(step_id))]
                if set(remote_log_files) >= {"controller", "stdout", "stderr"}:
                    break
                print "Fetching s3 logs..."
                time.sleep(retry_interval)
            for log_file in log_files:
                log_path = self.cluster.emr_step_log_filename(step_id, log_file)
                print "==================================="
                print "Dump EMR log file (step=%s): %s" % (step_id, log_file)
                print "Log Path : %s" % log_path
                print "==================================="
                if self.cluster.s3_list_files(log_path):
                    print self.cluster.emr_step_log(step_id, log_file=log_file)
                else:
                    print "Log file does not exist"
            return


class HiveRuntime(HadoopRuntime):
    def files_uploader(self, local_dir):
        return self.hdfs_upload_dir(local_dir)

    def hive_output_builder(self, output_name, output_obj):
        # TODO: refactor this method
        ps = self.settings
        glb_vars = ps.GlobalParam
        out_type = output_obj.types[0]
        if out_type.startswith("hive.table"):
            return "zetjob_%s_job%s_blk%s_OUTPUT_%s" % (
                glb_vars['userName'], glb_vars['jobId'], glb_vars['blockId'], output_name)
        elif out_type.startswith("hive.hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for hive, type must start with 'hive.table' or 'hive.hdfs'")

    def header_builder(self, hive_ns, uploaded_files, uploaded_jars):
        # Build Output Tables
        for output_name, output_obj in self.settings.Output._asdict().items():
            output_obj.val = self.hive_output_builder(output_name, output_obj)

        return "\n".join(
            itertools.chain(
                ["ADD FILE %s;" % f for f in uploaded_files],
                ["ADD JAR %s;" % f for f in uploaded_jars],
                ["set hivevar:MYNS = %s;" % hive_ns],
                ["set hivevar:PARAM_%s = %s;" % (k, v) for k, v in self.settings.Param._asdict().items()],
                ["set hivevar:INPUT_%s = %s;" % (k, v.val) for k, v in self.settings.Input._asdict().items()],
                ["set hivevar:OUTPUT_%s = %s;" % (k, v.val) for k, v in self.settings.Output._asdict().items()]))

    def clean_working_dir(self):
        self.hdfs_clean_working_dir()

    def generate_script(self, hive_script, target_filename=None):
        hive_ns = self.get_hive_namespace()

        # Upload files and UDF jars
        if 'FILE_DIR' in self.settings.Param._asdict():
            file_dir = self.settings.Param.FILE_DIR
            uploaded_files = self.files_uploader(file_dir.val)
        else:
            uploaded_files = []

        if 'UDF_DIR' in self.settings.Param._asdict():
            jar_dir = self.settings.Param.UDF_DIR
            uploaded_jars = self.files_uploader(jar_dir.val)
        else:
            uploaded_jars = []

        # Build Input, Output and Param
        header = self.header_builder(hive_ns, uploaded_files, uploaded_jars)
        if target_filename:
            import tempfile

            tmp_file = tempfile.NamedTemporaryFile(prefix="hive_generated_", suffix=".hql", delete=False)
            tmp_file.close()
            target_filename = tmp_file.name

        with open(hive_script, "r") as f, open(target_filename, "w+") as out_f:
            out_f.write("--------------------------\n")
            out_f.write("-- Header\n")
            out_f.write("--------------------------\n")
            out_f.write(header)
            out_f.write("\n")
            out_f.write("--------------------------\n")
            out_f.write("-- Main\n")
            out_f.write("--------------------------\n")
            out_f.write("\n")
            out_f.write(f.read())

        return target_filename

    def execute(self, hive_script, generated_hive_script=None):
        self.clean_working_dir()
        generated_hive_script = self.generate_script(hive_script, generated_hive_script)

        if cmd("beeline -u jdbc:hive2://%s:%s -n hive -p tiger -d org.apache.hive.jdbc.HiveDriver -f '%s' --verbose=true "
            % (self.settings.Param.HiveServer2_Host, self.settings.Param.HiveServer2_Port, generated_hive_script)) != 0:
            raise Exception("Failed to execute hive script : %s" % generated_hive_script)


class EmrHiveRuntime(EmrRuntime, HiveRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrHiveRuntime, self).__init__(spec_filename)

    def hive_output_builder(self, output_name, output_obj):
        # TODO : should refactor this function to base class
        ps = self.settings
        glb_vars = ps.GlobalParam
        out_type = output_obj.types[0]
        if out_type.startswith("hive.table"):
            return "zetjob_%s_job%s_blk%s_OUTPUT_%s" % (
                glb_vars['userName'], glb_vars['jobId'], glb_vars['blockId'], output_name)
        elif out_type.startswith("hive.hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        elif out_type.startswith("hive.s3"):
            return self.get_s3_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for hive, type must start with 'hive.table' or 'hive.hdfs' or 'hive.s3'")

    def files_uploader(self, local_dir):
        return self.s3_upload_dir(local_dir)

    def emr_execute_hive(self, s3_hive_script, dump_logfiles=None, dump_logfile_retry_count=1):
        step_ids = self.cluster.emr_execute_hive(self.get_emr_job_name(), [s3_hive_script])
        self.cluster.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        if not self.cluster.emr_wait_steps(step_ids):
            for step_id in step_ids:
                self.dump_logs(step_id, log_files=dump_logfiles, retry_count=dump_logfile_retry_count)
            raise Exception("EmrHiveRuntime: failed to execute steps")
        return step_ids

    def execute(self, main_hive_script, generated_hive_script=None, dump_logfiles=None, dump_logfile_retry_count=1):
        self.clean_working_dir()
        hive_script_local = self.generate_script(main_hive_script, generated_hive_script)

        s3_working_dir = self.get_s3_working_dir()
        hive_script_remote = os.path.join(s3_working_dir, os.path.basename(hive_script_local))
        hive_script_remote_full = s3_upload(self.s3_bucket, hive_script_local, hive_script_remote)
        print("========= Generated Hive Script =========")
        print(open(hive_script_local).read())
        print("=========================================")
        print("EmrHiveRuntime.execute()")
        return self.emr_execute_hive(hive_script_remote_full, dump_logfiles, dump_logfile_retry_count)


class EmrJarRuntime(EmrRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrJarRuntime, self).__init__(spec_filename)

    def execute(self, jar_path, args, dump_logfiles=None, dump_logfile_retry_count=1):
        s3_jar_path = s3_upload(self.s3_bucket, jar_path, self.get_s3_working_dir(jar_path))
        print("Uploading jar to s3 : %s -> %s" % (jar_path, s3_jar_path))

        step_ids = self.cluster.emr_execute_jar(job_name=self.get_emr_job_name(),
                                                s3_jar=s3_jar_path, args=args)
        self.cluster.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        if not self.cluster.emr_wait_steps(step_ids):
            for step_id in step_ids:
                self.dump_logs(step_id, log_files=dump_logfiles, retry_count=dump_logfile_retry_count)
            raise Exception("EmrJarRuntime : failed to execute steps")
        return step_ids


class PigRuntime(HadoopRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(PigRuntime, self).__init__(spec_filename)

    def files_uploader(self, local_dir):
        return self.hdfs_upload_dir(local_dir)

    def pig_output_builder(self, output_name, output_obj):
        # TODO: refactor this method
        out_type = output_obj.types[0]
        if out_type.startswith("pig.hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for pig, type must start with 'pig.hdfs'")

    def header_builder(self, uploaded_jars):
        # Build Output Tables
        for output_name, output_obj in self.settings.Output._asdict().items():
            output_obj.val = self.pig_output_builder(output_name, output_obj)

        return "\n".join(
            itertools.chain(
                ["%%declare PARAM_%s '%s'" % (k, v) for k, v in self.settings.Param._asdict().items()],
                ["%%declare INPUT_%s '%s'" % (k, v.val) for k, v in self.settings.Input._asdict().items()],
                ["%%declare OUTPUT_%s '%s'" % (k, v.val) for k, v in self.settings.Output._asdict().items()],
                ["REGISTER '%s';" % f for f in uploaded_jars]
            ))

    def generate_script(self, pig_script, target_filename=None):
        if 'UDF_DIR' in self.settings.Param._asdict():
            jar_dir = self.settings.Param.UDF_DIR
            uploaded_jars = self.files_uploader(jar_dir.val)
        else:
            uploaded_jars = []

        # Build Input, Output and Param
        header = self.header_builder(uploaded_jars)
        if target_filename:
            import tempfile

            tmp_file = tempfile.NamedTemporaryFile(prefix="pig_generated_", suffix=".hql", delete=False)
            tmp_file.close()
            target_filename = tmp_file.name

        with open(pig_script, "r") as f, open(target_filename, "w+") as out_f:
            out_f.write("/*************************\n")
            out_f.write(" * Header\n")
            out_f.write(" *************************/\n")
            out_f.write(header)
            out_f.write("\n")
            out_f.write("/*************************\n")
            out_f.write(" * Main\n")
            out_f.write(" *************************/\n")
            out_f.write("\n")
            out_f.write(f.read())
            out_f.write("\n")

        return target_filename

    def generate_pig_conf(self):
        ps = self.settings

        with open("/home/run/pig.properties", "a") as pf:
            pf.write("fs.default.name=%s\n" % ps.Param.hdfs_root)
            pf.write("yarn.resourcemanager.address=%s\n" % ps.Param.yarn_address)
            pf.write("yarn.resourcemanager.scheduler.address=%s\n" % ps.Param.yarn_scheduler_address)
        cmd("cat /home/run/pig.properties")

    def execute(self, pig_script):
        self.clean_working_dir()

        self.generate_pig_conf()
        generated_pig_script = self.generate_script(pig_script)
        print("========= Generated Pig Script =========")
        print(open(generated_pig_script).read())
        print("=========================================")
        print("EmrHiveRuntime.execute()")
        cmd("pig -x mapreduce -P /home/run/pig.properties %s" % generated_pig_script)


class EmrPigRuntime(EmrRuntime, PigRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrPigRuntime, self).__init__(spec_filename)

    def files_uploader(self, local_dir):
        return self.s3_upload_dir(local_dir)

    def pig_output_builder(self, output_name, output_obj):
        out_type = output_obj.types[0]
        if out_type.startswith("pig.hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        elif out_type.startswith("pig.s3"):
            return self.get_s3_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for pig, type must start with 'pig.hdfs' or 'pig.s3'")

    def emr_execute_pig(self, pig_filename, dump_logfiles=None, dump_logfile_retry_count=1):
        s3_pig_script = self.s3_upload(pig_filename)

        step_ids = self.cluster.emr_execute_pig(job_name=self.get_emr_job_name(), s3_pig_scripts=[s3_pig_script])
        self.cluster.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        if not self.cluster.emr_wait_steps(step_ids):
            for step_id in step_ids:
                self.dump_logs(step_id, log_files=dump_logfiles, retry_count=dump_logfile_retry_count)
            raise Exception("EmrPigRuntime: failed to execute steps")
        return step_ids

    def execute(self, pig_script, dump_logfiles=None, dump_logfile_retry_count=1):
        self.clean_working_dir()

        # TODO: upload S3 additional files

        generated_pig_script = self.generate_script(pig_script)
        print("========= Generated Pig Script =========")
        print(open(generated_pig_script).read())
        print("=========================================")
        print("EmrHiveRuntime.execute()")
        return self.emr_execute_pig(generated_pig_script, dump_logfiles, dump_logfile_retry_count)




# -*- coding: utf-8 -*-

"""
A series of Runtime.
"""

import itertools
import functools
from datacanvas.clusters import EmrCluster, QuboleCluster, EhcCluster, GenericHadoopCluster
from datacanvas.utils import *
from datacanvas.module import get_settings_from_file
from datacanvas import io_types
from datacanvas import version as datacanvas_version


class BasicRuntime(object):
    def __init__(self, spec_filename="spec.json"):
        self.settings = get_settings_from_file(spec_filename)

    def __repr__(self):
        return str(self.settings)

    @staticmethod
    def cmd(args, shell=False, verbose=True):
        if verbose:
            print("Execute External Command : '%s'" % args)
        ret = subprocess.call(args, shell=shell, env=os.environ.copy())
        if verbose:
            print("Exit with exit code = %d" % ret)
        return ret

    @staticmethod
    def exit(ret_code):
        sys.exit(ret_code)


def HadoopRuntime(*args, **kwargs):
    from warnings import warn
    warn("Use 'GenericHadoopRuntime' class! 'HadoopRuntime' is deprecated.")
    return GenericHadoopRuntime(*args, **kwargs)


class EmrRuntime(BasicRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrRuntime, self).__init__(spec_filename)
        self.grt = GenericHadoopRuntime()

    def get_s3_working_dir(self, path=""):
        return self.grt.get_working_dir(path)

    def get_emr_job_name(self):
        return self.grt.get_job_name()


class HiveRuntime(BasicRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(HiveRuntime, self).__init__(spec_filename)
        self.grt = GenericHadoopRuntime()

    def execute(self, hive_script, generated_hive_script=None):
        return self.grt.execute_hive(hive_script)


class EmrHiveRuntime(BasicRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrHiveRuntime, self).__init__(spec_filename)
        self.grt = GenericHadoopRuntime()

    def execute(self, main_hive_script, generated_hive_script=None, dump_logfiles=None, dump_logfile_retry_count=1):
        return self.grt.execute_hive(main_hive_script,
                                     logfiles=dump_logfiles,
                                     retry_count=dump_logfile_retry_count)


class PigRuntime(BasicRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(PigRuntime, self).__init__(spec_filename)
        self.grt = GenericHadoopRuntime()

    def execute(self, pig_script):
        self.grt.execute_pig(pig_script)


class EmrPigRuntime(BasicRuntime):
    def __init__(self, spec_filename="spec.json"):
        super(EmrPigRuntime, self).__init__(spec_filename)
        self.grt = GenericHadoopRuntime()

    def execute(self, pig_script, dump_logfiles=None, dump_logfile_retry_count=1):
        return self.grt.execute_pig(pig_script,
                                    logfiles=dump_logfiles,
                                    retry_count=dump_logfile_retry_count)


class ScriptBuilder(object):

    def __init__(self, settings, s3_working_root, hdfs_working_root, cluster_params=None):
        self.settings = settings
        self.s3_working_root = s3_working_root
        self.hdfs_working_root = hdfs_working_root
        if not cluster_params:
            cluster_params = {}
        self.cluster_params = cluster_params

    def get_hdfs_working_dir(self, dir_path=""):
        return s3join(self.hdfs_working_root, dir_path)

    def get_s3_working_dir(self, dir_path=""):
        return s3join(self.s3_working_root, dir_path)


class HiveScriptBuilder(ScriptBuilder):

    def __init__(self, settings, s3_working_root, hdfs_working_root, cluster_params, ignore_output_type_error=True):
        super(HiveScriptBuilder, self).__init__(settings, s3_working_root, hdfs_working_root, cluster_params)
        self.ignore_output_type_error = ignore_output_type_error

    def get_hive_namespace(self):
        ps = self.settings
        glb_vars = ps.GlobalParam
        return "zetjobns_{userName}_job{job_id}_blk{blk_id}".format(
            userName=glb_vars['userName'],
            job_id=glb_vars['jobId'],
            blk_id=glb_vars['blockId'])

    def get_hive_table(self, output_name):
        ps = self.settings
        glb_vars = ps.GlobalParam
        return "zetjob_{userName}_job{job_id}_blk{blk_id}_OUTPUT_{output_name}".format(
            userName=glb_vars['userName'],
            job_id=glb_vars['jobId'],
            blk_id=glb_vars['blockId'],
            output_name=output_name)

    def hive_output_builder(self, output_name, output_obj):
        out_type = output_obj.types[0]
        # TODO: refactor types
        if out_type.startswith("hive.table"):
            return self.get_hive_table(output_name)
        elif out_type.startswith("hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        elif out_type.startswith("s3"):
            return self.get_s3_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for hive")

    def hive_create_output(self, output_name, output_obj):
        cluster_params = self.cluster_params

        out_type = output_obj.types[0]
        if io_types.is_type_of("s3", out_type):
            aws_key = cluster_params.get("accessKey", "")
            aws_sec = cluster_params.get("accessSecret", "")
            return io_types.DS_S3(URL=self.get_s3_working_dir("OUTPUT_%s" % output_name),
                                  aws_key=aws_key,
                                  aws_security=aws_sec)
        elif io_types.is_type_of("hdfs", out_type) or io_types.is_type_of("datasource.hdfs", out_type):
            return io_types.DS_HDFS(URL=self.get_hdfs_working_dir("OUTPUT_%s" % output_name))
        elif io_types.is_type_of("hive", out_type) or io_types.is_type_of("datasource.hive", out_type):
            hs2_host = cluster_params.get("hive_server2_host", "")
            hs2_port = cluster_params.get("hive_server2_port", "")
            if hs2_host and hs2_port:
                return io_types.DS_Hive(URL=self.get_hive_table("OUTPUT_%s" % output_name),
                                        meta_server=hs2_host,
                                        meta_port=hs2_port)
            else:
                return io_types.DS_Hive(URL=self.get_hive_table("OUTPUT_%s" % output_name))
        elif io_types.is_type_of("datasource", out_type):
            # TODO: default as a 'hive' table?
            return io_types.DS_Hive(URL=self.get_hive_table("OUTPUT_%s" % output_name))
        else:
            if self.ignore_output_type_error:
                return None
            else:
                raise ValueError("Invalid type for hive")

    def header_builder(self, hive_ns, uploaded_files, uploaded_jars, extra_vars=None):
        if not extra_vars:
            extra_vars = {}

        # Build Output Tables
        for output_name, output_obj in self.settings.Output._asdict().items():
            output_obj.val = self.hive_create_output(output_name, output_obj)

        def _get_io_val(obj):
            if 'URL' in obj:
                return obj['URL']
            else:
                # TODO:
                if self.ignore_output_type_error:
                    return ""
                else:
                    raise Exception("TODO: ERROR")

        return "\n".join(
            itertools.chain(
                ["ADD FILE %s;" % f for f in uploaded_files],
                ["ADD JAR %s;" % f for f in uploaded_jars],
                ["set hivevar:MYNS = %s;" % hive_ns],
                ["set hivevar:PARAM_%s = %s;" % (k, v) for k, v in self.settings.Param._asdict().items() if v.is_primitive],
                ["set hivevar:INPUT_%s = %s;" % (k, _get_io_val(v.val)) for k, v in self.settings.Input._asdict().items()],
                ["set hivevar:OUTPUT_%s = %s;" % (k, _get_io_val(v.val)) for k, v in self.settings.Output._asdict().items() if v.val],
                ["set hivevar:%s = %s;" % (k, v) for k, v in extra_vars.items()]))

    def generate_script(self, hive_script, uploaded_files, uploaded_jars, extra_vars=None):
        hive_ns = self.get_hive_namespace()

        # Build Input, Output and Param
        header = self.header_builder(hive_ns, uploaded_files, uploaded_jars, extra_vars)

        import tempfile
        tmp_file = tempfile.NamedTemporaryFile(prefix="hive_generated_", suffix=".hql", delete=False)
        tmp_file.close()
        target_filename = tmp_file.name

        with open(hive_script, "r") as f, open(target_filename, "w+") as out_f:
            out_f.write("--------------------------\n")
            out_f.write("-- Header\n")
            out_f.write("--------------------------\n")
            out_f.write(header)
            out_f.write("\n")
            out_f.write("--------------------------\n")
            out_f.write("-- Main\n")
            out_f.write("--------------------------\n")
            out_f.write("\n")
            out_f.write(f.read())

        return target_filename


class PigScriptBuilder(ScriptBuilder):

    def __init__(self, settings, s3_working_root, hdfs_working_root, cluster_params):
        super(PigScriptBuilder, self).__init__(settings, s3_working_root, hdfs_working_root, cluster_params)

    def pig_output_builder(self, output_name, output_obj):
        out_type = output_obj.types[0]
        # TODO: refactor types
        if out_type.startswith("hdfs"):
            return self.get_hdfs_working_dir("OUTPUT_%s" % output_name)
        elif out_type.startswith("s3"):
            return self.get_s3_working_dir("OUTPUT_%s" % output_name)
        else:
            raise ValueError("Invalid type for hive, type must start with 'hive.table' or 'hive.hdfs' or 'hive.s3'")

    def pig_create_output(self, output_name, output_obj):
        cluster_params = self.cluster_params

        out_type = output_obj.types[0]
        if io_types.is_type_of("s3", out_type):
            aws_key = cluster_params.get("accessKey", None)
            aws_sec = cluster_params.get("accessSecret", None)
            return io_types.DS_S3(URL=self.get_s3_working_dir("OUTPUT_%s" % output_name),
                                  aws_key=aws_key,
                                  aws_security=aws_sec)
        elif io_types.is_type_of("hdfs", out_type):
            return io_types.DS_HDFS(URL=self.get_hdfs_working_dir("OUTPUT_%s" % output_name))
        else:
            raise ValueError("Invalid type for hive")

    def header_builder(self, uploaded_jars, extra_vars=None):
        if not extra_vars:
            extra_vars = {}

        # Build Output Tables
        for output_name, output_obj in self.settings.Output._asdict().items():
            output_obj.val = self.pig_create_output(output_name, output_obj)

        return "\n".join(
            itertools.chain(
                ["%%declare PARAM_%s '%s'" % (k, v)
                 for k, v in self.settings.Param._asdict().items()
                 if v.is_primitive],
                ["%%declare INPUT_%s '%s'" % (k, v.val)
                 for k, v in self.settings.Input._asdict().items()],
                ["%%declare OUTPUT_%s '%s'" % (k, v.val)
                 for k, v in self.settings.Output._asdict().items()],
                ["%%declare %s '%s'" % (k, v.val)
                 for k, v in extra_vars.items()],
                ["REGISTER '%s';" % f
                 for f in uploaded_jars]
            ))

    def generate_script(self, pig_script, uploaded_jars, extra_vars=None):

        # Build Input, Output and Param
        header = self.header_builder(uploaded_jars, extra_vars)

        import tempfile
        tmp_file = tempfile.NamedTemporaryFile(prefix="pig_generated_", suffix=".pig", delete=False)
        tmp_file.close()
        target_filename = tmp_file.name

        with open(pig_script, "r") as f, open(target_filename, "w+") as out_f:
            out_f.write("/*************************\n")
            out_f.write(" * Header\n")
            out_f.write(" *************************/\n")
            out_f.write(header)
            out_f.write("\n")
            out_f.write("/*************************\n")
            out_f.write(" * Main\n")
            out_f.write(" *************************/\n")
            out_f.write("\n")
            out_f.write(f.read())
            out_f.write("\n")

        return target_filename


class GenericHadoopRuntime(BasicRuntime):

    def __init__(self, cluster_var_name="cluster"):
        super(GenericHadoopRuntime, self).__init__()
        self.hadoop_type = None
        self.cluster = None
        self.working_root = None
        self.hdfs_working_root = None
        self.s3_working_root = None
        self.global_params = self.settings.GlobalParam
        self.cluster_params = None

        param_dict = self.settings.Param._asdict()
        if (cluster_var_name in param_dict) and param_dict[cluster_var_name].is_cluster:
            cluster_type = self._get_cluster_type(cluster_var_name)
            print param_dict.get(cluster_var_name).type
            print "=================================================="
            print "Use cluster var :: '%s'" % cluster_var_name
            print "           type :: '%s'" % cluster_type
            print "=================================================="
            self.switch_hadoop_env(cluster_type, cluster_var_name)

    def _get_cluster_type(self, cluster_var_name):
        cparam = self.settings.Param._asdict().get(cluster_var_name).val
        return cparam["Type"]

    def _get_cluster_params(self, cluster_var_name):
        cparam = self.settings.Param._asdict().get(cluster_var_name).val
        cluster_params = {p["Name"]: p.get("Val", None) for p in cparam["Parameters"]}
        return cluster_params

    def switch_hadoop_env(self, hadoop_type, cluster_var_name="cluster", extra_env_vars=None):
        print "Switching to Hadoop type = '%s'" % hadoop_type
        print "pyDataCanvas version : '%s'" % datacanvas_version
        cluster_params = self._get_cluster_params(cluster_var_name)
        self.cluster_params = cluster_params

        if hadoop_type in ["EMR", "EMR_SPOT"]:
            self.hadoop_type = hadoop_type
            self.cluster = EmrCluster(aws_region=cluster_params["region"],
                                      aws_key=cluster_params["accessKey"],
                                      aws_secret=cluster_params["accessSecret"],
                                      jobflow_id=cluster_params["jobFlowId"])
            self.cluster.prepare(hadoop_type, **cluster_params)
            self.working_root = self.cluster.get_working_root(cluster_params, self.global_params)
            self.s3_working_root = self.working_root
            self.hdfs_working_root = "/"
        elif hadoop_type in ["CDH4", "CDH5"]:
            self.hadoop_type = hadoop_type
            self.cluster = GenericHadoopCluster(**cluster_params)
            self.cluster.prepare(hadoop_type, **cluster_params)
            self.working_root = self.cluster.get_working_root(cluster_params, self.global_params)
            self.s3_working_root = None
            self.hdfs_working_root = cluster_params["hdfs_root"]
        elif hadoop_type in ["QUBOLE"]:
            self.hadoop_type = hadoop_type
            self.cluster = QuboleCluster(qubole_api_token=cluster_params["qubole_api_token"],
                                         qubole_cluster_name=cluster_params["qubole_cluster_name"],
                                         qubole_tags=cluster_params["qubole_tags"])
            self.cluster.prepare(hadoop_type, **cluster_params)
            self.working_root = self.cluster.get_working_root(cluster_params, self.global_params)
            self.s3_working_root = self.working_root
            self.hdfs_working_root = "/"
        elif hadoop_type in ["EHC"]:
            self.hadoop_type = hadoop_type
            self.cluster = EhcCluster(ehc_token=cluster_params["ehc_token"],
                                      ehc_id=cluster_params["ehc_id"])
            self.cluster.prepare(hadoop_type, **cluster_params)
            self.working_root = self.cluster.get_working_root(cluster_params, self.global_params)
            self.s3_working_root = self.working_root
            self.hdfs_working_root = "/"
        else:
            raise Exception("Do NOT support hadoop_type '%s'" % hadoop_type)

        self.cluster.clean_working_dir(self.working_root)
        print self.working_root

    def get_working_dir(self, path=""):
        if not self.working_root:
            raise ValueError("Did NOT define 'working_root'!")

        return s3join(self.working_root, path)

    def get_job_name(self):
        ps = self.settings
        glb_vars = ps.GlobalParam
        return os.path.join('zetjob', glb_vars['userName'],
                            "job%s" % glb_vars['jobId'], "blk%s" % glb_vars['blockId'])

    def execute_jar(self, jar_path, jar_args, main_class="", *args, **kwargs):
        job_name = self.get_job_name()

        remote_jar_path = self.cluster.prepare_working_file(self.working_root, jar_path)
        return self.cluster.execute_jar(job_name=job_name, jar_path=remote_jar_path,
                                        jar_args=jar_args, main_class=main_class,
                                        *args, **kwargs)

    def execute_hive_filename(self, hive_main_filename, extra_vars=None, *args, **kwargs):
        job_name = self.get_job_name()

        hb = HiveScriptBuilder(self.settings, s3_working_root=self.s3_working_root,
                               hdfs_working_root=self.hdfs_working_root,
                               cluster_params=self.cluster_params)
        generated_hql = hb.generate_script(hive_main_filename, [], [], extra_vars=extra_vars)
        remote_hive_script = self.cluster.prepare_working_file(self.working_root, generated_hql)
        ret = self.cluster.execute_hive(job_name=job_name,
                                        hive_script=remote_hive_script,
                                        *args, **kwargs)
        return ret

    def execute_pig_filename(self, pig_main_filename, extra_vars=None, *args, **kwargs):
        job_name = self.get_job_name()

        pb = PigScriptBuilder(self.settings, s3_working_root=self.s3_working_root,
                              hdfs_working_root=self.hdfs_working_root,
                              cluster_params=self.cluster_params)
        generated_pig = pb.generate_script(pig_main_filename, [], extra_vars=extra_vars)
        remote_pig_script = self.cluster.prepare_working_file(self.working_root, generated_pig)
        ret = self.cluster.execute_pig(job_name=job_name,
                                       pig_script=remote_pig_script,
                                       *args, **kwargs)
        return ret


##############
# Decorators #
##############

class DataCanvas(object):
    """DataCanvas"""

    def __init__(self, name):
        self._name = name
        self._graph = []
        self._rt = None

    def basic_runtime(self, spec_json="spec.json"):
        def decorator(method):
            rt = BasicRuntime(spec_filename=spec_json)
            params = rt.settings.Param
            inputs = rt.settings.Input
            outputs = rt.settings.Output

            @functools.wraps(method)
            def wrapper(_rt=rt, _params=params, _inputs=inputs, _outputs=outputs):
                print rt
                method(_rt, _params, _inputs, _outputs)

            self._graph.append(wrapper)
            return wrapper

        return decorator

    def hadoop_runtime(self, spec_json="spec.json"):
        def decorator(method):
            rt = GenericHadoopRuntime(cluster_var_name="cluster")
            params = rt.settings.Param
            inputs = rt.settings.Input
            outputs = rt.settings.Output

            @functools.wraps(method)
            def wrapper(_rt=rt, _params=params, _inputs=inputs, _outputs=outputs):
                print rt
                method(_rt, _params, _inputs, _outputs)

            self._graph.append(wrapper)
            return wrapper

        return decorator

    def run(self):
        for m in self._graph:
            m()



# -*- coding: utf-8 -*-

from .version import version

__version__ = version
__author__ = "xiaolin"

from clusters import EmrCluster
from log_parser import parse_syslog, parse_s3_stderr
from module import Input, Output, Param
from runtime import DatacanvasRuntime, HadoopRuntime, PigRuntime, HiveRuntime, \
    EmrRuntime, EmrHiveRuntime, EmrJarRuntime, EmrPigRuntime
from utils import cmd



# -*- coding: utf-8 -*-

import os
import re
import sys
import json
import types
from collections import namedtuple
from datacanvas.io_types import load_io_obj, BaseIO
from datacanvas.utils import mask_key


class Input(str):
    def __new__(cls, x, _types):
        return str.__new__(cls, x)

    def __init__(self, x, _types):
        super(Input, self).__init__()
        self.x = x
        self._types = _types

    def __repr__(self):
        return str(self.x)

    def __str__(self):
        return str(self.x)

    def as_first_line(self):
        with open(self.x, "r") as f:
            return f.readline().rstrip()

    def as_whole(self):
        with open(self.x, "r") as f:
            return f.read()

    def as_file(self, mode="r"):
        return open(self.x, mode)

    def as_datasource(self, mode="r"):
        ds = json.loads(open(self.x, mode).read())
        return ds

    def as_obj(self):
        ds = load_io_obj(self.x)
        return ds

    def as_raw(self):
        return str(self.x)

    @property
    def val(self):
        """Unboxing Input depends on its types."""
        if "any" in self._types:
            return self.as_raw()
        elif any([re.match(r"raw.*", t) for t in self._types]):
            return self.as_raw()
        else:
            return self.as_obj()

    @property
    def types(self):
        return self._types


class Output(str):
    def __new__(cls, x, _types):
        return str.__new__(cls, x)

    def __init__(self, x, _types):
        super(Output, self).__init__()
        self.x = x
        self._types = _types

    def __repr__(self):
        return str(self.x)

    def __str__(self):
        return str(self.x)

    def as_first_line(self):
        with open(self.x, "r") as f:
            return f.readline().rstrip()

    def as_whole(self):
        with open(self.x, "r") as f:
            return f.read()

    def as_file(self, mode="r"):
        return open(self.x, mode)

    def as_obj(self):
        ds = load_io_obj(self.x)
        return ds

    def as_raw(self):
        return str(self.x)

    @property
    def val(self):
        """Unboxing Input depends on its types."""
        if "any" in self._types:
            return self.as_raw()
        elif any([re.match(r"raw.*", t) for t in self._types]):
            return self.as_raw()
        else:
            return load_io_obj(self.x)

    @val.setter
    def val(self, value):
        with open(self.x, "w+") as f:
            if "any" in self._types or any([re.match(r"raw.*", t) for t in self._types]):
                return

            if isinstance(value, BaseIO):
                f.write(json.dumps(value))
            else:
                f.write(value)

    @property
    def types(self):
        return self._types


class Param(str):
    def __new__(cls, x, typeinfo):
        return str.__new__(cls, x)

    def __init__(self, x, typeinfo):
        super(Param, self).__init__()
        self._x = x
        self._typeinfo = typeinfo

    def __repr__(self):
        if self.is_cluster:
            return self.show()
        else:
            return str(self._x)

    def __str__(self):
        if self.is_cluster:
            return self.show()
        else:
            return str(self._x)

    def show(self, mask_keys=True):
        def get_safe_cluster_param(cp):
            security_mask_names = ['accessKey', 'accessSecret', 'qubole_api_token', 'subscriptionId']
            if mask_keys and cp['Name'] in security_mask_names and 'Val' in cp:
                cp['Val'] = mask_key(cp['Val'])
                return cp
            return cp

        o = self.val
        if isinstance(o, dict):
            if self.is_cluster:
                cluster_params = o['Parameters']
                o['Parameters'] = [get_safe_cluster_param(cp) for cp in cluster_params]
                return json.dumps(o)
        else:
            return str(self._x)

    @property
    def type(self):
        return self._typeinfo['Type']

    @property
    def is_primitive(self):
        return self._typeinfo['Type'] not in ["cluster"]

    @property
    def is_cluster(self):
        return self._typeinfo['Type'] in ["cluster"]

    @property
    def val(self):
        type_handler = {
            "string": lambda x: x,
            "float": lambda x: float(x),
            "integer": lambda x: int(x),
            "enum": lambda x: x,
            "cluster": lambda x: json.loads(x),
            "file": read_whole_file
        }
        param_type = self._typeinfo['Type']
        if param_type in type_handler:
            return type_handler[param_type](self._x)
        else:
            return self._x


def read_whole_file(filename):
    with open(filename, "r") as f:
        return f.read()


def gettype(name):
    type_map = {
        "string": "str",
        "integer": "int",
        "float": "float",
        "enum": "str",
        "file": "str"
    }
    if name not in type_map:
        raise ValueError(name)
    t = __builtins__.get(type_map[name], types.StringType)
    if isinstance(t, type):
        return t
    raise ValueError(name)


def input_output_builder(spec_input, spec_output):
    params = dict(arg.split("=") for arg in sys.argv[1:])
    if not all(k in params for k in spec_input.keys()):
        raise ValueError("Missing input parameters")
    if not all(k in params for k in spec_output.keys()):
        raise ValueError("Missing output parameters")

    inputSettings = namedtuple('InputSettings', spec_input.keys())
    in_params = {in_k: Input(params[in_k], in_type) for in_k, in_type in spec_input.items()}
    input_settings = inputSettings(**in_params)

    OutputSettings = namedtuple('OutputSettings', spec_output.keys())
    out_params = {out_k: Output(params[out_k], out_type) for out_k, out_type in spec_output.items()}
    output_settings = OutputSettings(**out_params)

    return input_settings, output_settings


def param_builder(spec_param, param_json):
    def get_param(k):
        if k in param_json:
            return param_json[k]['Val']
        else:
            return spec_param[k]['Default']

    ParamSettings = namedtuple('ParamSettings', spec_param.keys())
    param_dict = {k: Param(get_param(k), v) for k, v in spec_param.items()}
    env_settings = ParamSettings(**param_dict)
    return env_settings


def global_param_builder(param_json):
    return {k: v['Val'] for k, v in param_json.items()}


def get_settings(spec_json):
    moderate_keys = ['Name', 'Param', 'Input', 'Output', 'Cmd', 'Description']
    if not all(k in spec_json for k in moderate_keys):
        raise ValueError("One of param from %s may not exist in 'spec.json'" % str(moderate_keys))

    # TODO: condition for appending 'GlobalParam'
    moderate_keys.append('GlobalParam')
    ModuleSetting = namedtuple('ModuleSetting', moderate_keys)

    # Load parameters
    param_json = get_json_file(os.getenv("ZETRT"))

    param = param_builder(spec_json['Param'], param_json['PARAM'])
    json_input, json_output = input_output_builder(spec_json['Input'], spec_json['Output'])

    # TODO:
    global_param = global_param_builder(param_json['GLOBAL_PARAM'])
    settings = ModuleSetting(Name=spec_json['Name'], Description=spec_json['Description'], Param=param,
                             Input=json_input, Output=json_output, Cmd=spec_json['Cmd'], GlobalParam=global_param)
    return settings


def get_json_file(filename):
    with open(filename, "r") as f:
        return json.load(f)


def get_settings_from_file(filename):
    with open(filename, "r") as f:
        return get_settings(json.load(f))


def get_settings_from_string(spec_json_str):
    print(json.loads(spec_json_str))
    return get_settings(json.loads(spec_json_str))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import time
import requests
from datacanvas.utils import url_path_join


class EhcClient(object):
    def __init__(self, ehc_token, ehc_id, ehc_api_root="https://ehc.datacanvas.io:4430"):
        self.ehc_token = ehc_token
        self.ehc_id = ehc_id
        self.api_url = ehc_api_root

    def _post_command(self, cmd_path, cmd_type, **cmd_args):
        headers = {
            "Content-Type": "application/json",
            "X-AUTH-TOKEN": self.ehc_token
        }
        cmd_body = {
            "ehc_setting": {
                "id": self.ehc_id
            },
            "command_setting": {
                "type": cmd_type,
                "conf": cmd_args
            }
        }

        return requests.post(url_path_join(self.api_url, cmd_path),
                             headers=headers,
                             data=json.dumps(cmd_body))

    def _get_command(self, query_type, cmd_id):
        headers = {
            "Content-Type": "application/json",
            "X-AUTH-TOKEN": self.ehc_token
        }
        cmd_body = {
            "command_setting": {
                "id": cmd_id
            }
        }

        return requests.post(url_path_join(self.api_url, "/api/ehc/command/", query_type),
                             headers=headers,
                             data=json.dumps(cmd_body))

    def _get_user(self):
        headers = {
            "X-AUTH-TOKEN": self.ehc_token
        }
        return requests.get(url_path_join(self.api_url, "/api/ehc/user"),
                            headers=headers)

    def create_shell_command(self, cmd):
        req_handler = "/api/ehc/command/submit"
        r = self._post_command(req_handler, "shell", inline=cmd)
        if r.status_code != 200:
            raise ServiceError("Failed to request '%s'" % req_handler)
        return r.json()

    def create_hadoop_jar_command(self):
        pass

    def create_hive_command(self, query):
        pass

    def get_command_status(self, cmd_id):
        r = self._get_command(query_type="status", cmd_id=cmd_id)
        if r.status_code != 200:
            raise ServiceError("Failed to get command status")
        return r.json()

    def get_command_results(self, cmd_id):
        r = self._get_command(query_type="results", cmd_id=cmd_id)
        if r.status_code != 200:
            raise ServiceError("Failed to get command results")
        return r.json()

    def get_user(self):
        r = self._get_user()
        if r.status_code != 200:
            raise ServiceError("Failed to get user.")
        return r.json()

    def wait_command(self, cmd_id, interval=1, verbose=True):
        while True:
            r = self.get_command_status(cmd_id)
            job_status = r['status']
            if verbose:
                print "EHC.Command(ehc_id = '%s', cmd_id = '%s') Status is '%s'" % (self.ehc_id, cmd_id, job_status)
            if job_status in ["running", "prep"]:
                time.sleep(interval)
                continue
            return job_status


class BaseError(Exception):
    """
    General Ehc Client error
    """
    def __init__(self, reason, *args):
        super(BaseError, self).__init__(reason, *args)
        self.reason = reason

    def __repr__(self):
        return 'EhcClientError: %s' % self.reason

    def __str__(self):
        return 'EhcClientError: %s' % self.reason


class ServiceError(BaseError):
    """
    Error when request to EHC service.
    """
    def __init__(self, reason, message):
        super(ServiceError, self).__init__(reason, message)
        self.message = message

    def __repr__(self):
        return 'ServiceError: %s' % self.reason

    def __str__(self):
        return 'ServiceError: %s' % self.reason



# -*- coding: utf-8 -*-

import os
import subprocess
import time
import boto
import boto.s3
import boto.s3.key
import boto.emr
from boto.emr.step import HiveStep, JarStep, PigStep
from itertools import izip
from datacanvas.utils import cached_property, url_path_join, s3join, \
    s3parse, s3_upload, pprint_aws_obj, prepare_hadoop_conf, \
    preprocess_cluster_envs, percent_cb
from urlparse import urlparse
import abc
import qds_sdk.qubole
import qds_sdk.commands
import qds_sdk.cluster
from datacanvas.ehc_client import EhcClient


class BaseCluster(object):
    """A BaseCluster class define abstract methods for some type of cluster"""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        pass

    def check_variables(self):
        pass

    @abc.abstractmethod
    def prepare(self, hadoop_type, **cluster_kws):
        pass

    @abc.abstractmethod
    def get_working_root(self, cluster_params, global_params):
        pass

    @abc.abstractmethod
    def upload_working_file(self, working_root, file_path):
        pass

    @abc.abstractmethod
    def upload_working_dir(self, working_root, local_path):
        pass

    @abc.abstractmethod
    def clean_working_dir(self, dir_path):
        pass

    @abc.abstractmethod
    def prepare_working_file(self, working_root, file_path):
        pass

    @abc.abstractmethod
    def execute_jar(self, job_name, jar_path, jar_args, main_class, *args, **kwargs):
        pass

    # TODO: execute_stream_jar
    def execute_stream_jar(self, job_name, jar_file, jar_args, *args, **kwargs):
        pass

    @abc.abstractmethod
    def execute_hive(self, job_name, hive_script, *args, **kwargs):
        pass

    @abc.abstractmethod
    def execute_pig(self, job_name, pig_script, *args, **kwargs):
        pass


class EmrBaseCluster(object):
    """A class to access EMR cluster resources.

    Examples:
        aws_key = "XXXXXXXXXXXXXXXXXXXX"
        aws_sec = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        job_id  = "j-xxxxxxxxxxxx"
        step_id = "s-xxxxxxxxxxxx"

        emr_cluster = EmrCluster(job_id, aws_region, aws_key, aws_sec)

        # Get 'stdout' log file
        emr_cluster.get_step_log(step_id)

        # Get 'stderr' log file
        emr_cluster.get_step_log(step_id， logfile="stderr")
    """

    def __init__(self, **kwargs):
        self.aws_region = None
        self.aws_key = None
        self.aws_secret = None
        self.jobflow_id = None
        self._kwargs = kwargs

    @cached_property
    def emr_conn(self):
        return boto.connect_emr(region=self.aws_region,
                                aws_access_key_id=self.aws_key,
                                aws_secret_access_key=self.aws_secret)

    @cached_property
    def s3_conn(self):
        return boto.connect_s3(self.aws_key, self.aws_secret)

    @cached_property
    def emr_info(self):
        return self.emr_conn.describe_cluster(self.jobflow_id)

    @property
    def emr_steps(self):
        steps = self.emr_conn.list_steps(self.jobflow_id)

        all_steps = [{"index": len(steps.steps) - si, "id": se.id, "name": se.name}
                     for si, se in enumerate(steps.steps)]
        steps_dict = {s["id"]: s for s in all_steps}
        return steps_dict

    def s3_list_files(self, path):
        pr = s3parse(path)
        if pr.scheme not in ["s3", "s3n"]:
            raise ValueError("Not valid s3 path: '%s'" % path)
        bucket = self.s3_conn.get_bucket(pr.netloc)
        prefix_path = pr.path[1:]
        return [k.name.encode('utf-8') for k in bucket.list(prefix=prefix_path, delimiter="/")]

    def s3_get_file(self, s3_path):
        pr = s3parse(s3_path)
        if pr.scheme not in ["s3", "s3n"]:
            raise ValueError("Not valid s3 path: '%s'" % s3_path)
        bucket = self.s3_conn.get_bucket(pr.netloc)
        return bucket.get_key(pr.path)

    def s3_delete_file(self, s3_path):
        pr = s3parse(s3_path)
        if pr.scheme not in ["s3", "s3n"]:
            raise ValueError("Not valid s3 path: '%s'" % s3_path)
        bucket = self.s3_conn.get_bucket(pr.netloc)
        prefix_path = pr.path[1:]
        for key in bucket.list(prefix=prefix_path):
            key.delete()
        return True

    def s3_sync(self, src_dir, dest_dir):
        pr = s3parse(src_dir)
        bucket = self.s3_conn.get_bucket(pr.netloc)
        for f in self.s3_list_files(src_dir):
            src_fn = s3join("s3://"+pr.netloc, f)
            src_pr = s3parse(src_fn)
            k = bucket.get_key(src_pr.path)

            print "Downloading '%s'" % src_fn
            dest_fn = os.path.join(dest_dir, os.path.basename(src_pr.path))
            k.get_contents_to_filename(dest_fn, cb=percent_cb, num_cb=1000)
            print ""

        return dest_dir

    def dump_logs(self, step_id, log_files=None, retry_count=1, retry_interval=60, extra_wait_count=5):
        if not log_files:
            return

        while retry_count > 0:
            remote_log_files = [os.path.basename(i)
                                for i in self.s3_list_files(self.emr_step_log_filename(step_id))]
            intersection_files = set(log_files) & set(remote_log_files)
            if not intersection_files:
                print "Seems logs are not ready on s3, retry..."
                retry_count -= 1
                time.sleep(retry_interval)
                continue
            for i in range(extra_wait_count):
                remote_log_files = [os.path.basename(i)
                                    for i in self.s3_list_files(self.emr_step_log_filename(step_id))]
                if set(remote_log_files) >= {"controller", "stdout", "stderr"}:
                    break
                print "Fetching s3 logs..."
                time.sleep(retry_interval)
            for log_file in log_files:
                log_path = self.emr_step_log_filename(step_id, log_file)
                print "==================================="
                print "Dump EMR log file (step=%s): %s" % (step_id, log_file)
                print "Log Path : %s" % log_path
                print "==================================="
                if self.s3_list_files(log_path):
                    print self.emr_step_log(step_id, log_file=log_file)
                else:
                    print "Log file does not exist"
            return

    def emr_execute_jar(self, job_name, s3_jar_path, jar_args, main_class="", action_on_failure='CONTINUE'):
        steps = [JarStep(name=job_name, jar=s3_jar_path,
                         main_class=main_class, step_args=jar_args,
                         action_on_failure=action_on_failure)]
        ret_steps = self.emr_conn.add_jobflow_steps(self.jobflow_id, steps=steps)
        step_ids = [s.value for s in ret_steps.stepids]
        return step_ids

    def emr_execute_hive(self, job_name, s3_hive_script):
        hive_step = HiveStep(name=job_name, hive_file=s3_hive_script)
        hive_step.action_on_failure = 'CONTINUE'
        ret_steps = self.emr_conn.add_jobflow_steps(self.jobflow_id, steps=[hive_step])
        step_ids = [s.value for s in ret_steps.stepids]
        return step_ids

    def emr_execute_pig(self, job_name, s3_pig_script):
        pig_step = PigStep(name=job_name, pig_file=s3_pig_script)
        pig_step.action_on_failure = 'CONTINUE'
        ret_steps = self.emr_conn.add_jobflow_steps(self.jobflow_id, steps=[pig_step])
        step_ids = [s.value for s in ret_steps.stepids]
        return step_ids

    def emr_describe_steps(self, step_ids):
        for sid in step_ids:
            print "==================================="
            print "Summary for step: %s" % sid
            print "==================================="
            step = self.emr_conn.describe_step(self.jobflow_id, sid)
            # print pprint_aws_obj(step)

    def emr_wait_steps(self, stepids):
        for step_id in stepids:
            while True:
                ret = self.emr_conn.describe_step(self.jobflow_id, step_id)
                step_state = ret.status.state
                print "Wait EMR jobflow step: jobflow_id='%s' step_id='%s' state='%s'" % \
                      (self.jobflow_id, step_id, step_state)
                if step_state in ["PENDING", "RUNNING", "CONTINUE"]:
                    time.sleep(10)
                    continue
                elif step_state in ['COMPLETED']:
                    break
                else:
                    return False
        return True

    def s3_upload(self, local_path, remote_s3_path, recursive=False):
        if os.path.isdir(local_path) and not recursive:
            raise ValueError("Can not upload directory, set 'recursive=True'")

        s3_pr = s3parse(remote_s3_path)
        if s3_pr.scheme not in ["s3", "s3n"]:
            raise ValueError("Not valid s3 path: '%s'" % remote_s3_path)
        bucket = self.s3_conn.get_bucket(s3_pr.netloc)

        upload_files = []
        remote_files = []
        if os.path.isfile(local_path):
            upload_files = [local_path]
            remote_files = [s3join(remote_s3_path, os.path.basename(local_path))]
        elif os.path.isdir(local_path):
            upload_files = [os.path.join(dirpath, fn)
                            for dirpath, dirnames, filenames in os.walk(local_path)
                            for fn in filenames]
            remote_files = [s3join(remote_s3_path, fn_local) for fn_local in upload_files]

        for fn_local, fn_remote in izip(upload_files, remote_files):
            s3_upload(bucket, fn_local, fn_remote)

        return remote_files

    def emr_step_log_filename(self, step_id, log_file=None):
        if not log_file:
            log_file = "/"
        return url_path_join(self.emr_info.loguri, self.jobflow_id,
                             "steps", str(self.emr_steps[step_id]["index"]), log_file)

    def emr_step_log(self, step_id, log_file="stdout", local_file=None):
        fn = self.emr_step_log_filename(step_id, log_file)

        pr = s3parse(fn)
        bucket = self.s3_conn.get_bucket(pr.netloc)
        k = boto.s3.key.Key(bucket)
        k.key = pr.path
        if not local_file:
            return k.get_contents_as_string()
        else:
            # TODO: Not implemented
            raise Exception("Not implemented")

    def _wait_steps(self, step_ids, **kwargs):
        log_files = kwargs.get("log_files", ["controller", "stderr", "stdout"])
        retry_count = kwargs.get("retry_count", 10)
        retry_interval = kwargs.get("retry_interval", 60)
        extra_wait_count = kwargs.get("extra_wait_count", 5)
        if not self.emr_wait_steps(step_ids):
            for step_id in step_ids:
                self.dump_logs(step_id, log_files=log_files, retry_count=retry_count,
                               retry_interval=retry_interval,
                               extra_wait_count=extra_wait_count)
            raise Exception("EmrCluster: failed to execute steps")


class EmrMixin(BaseCluster):

    def prepare(self, hadoop_type, **cluster_kws):
        if hadoop_type not in ["EMR", "EMR_SPOT"]:
            raise Exception("EmrCluster :: Can NOT prepare '%s'" % hadoop_type)

        self.aws_region = cluster_kws["region"]
        self.aws_key = cluster_kws["accessKey"]
        self.aws_secret = cluster_kws["accessSecret"]
        self.jobflow_id = cluster_kws.get("jobflow_id", None)

    def get_working_root(self, cluster_params, global_params):
        remote_path = "s3n://{working_bucket}/zetjob/{username}/job{job_id}/blk{blk_id}/".format(
            working_bucket=cluster_params["S3_BUCKET"],
            username=global_params["userName"],
            job_id=global_params["jobId"],
            blk_id=global_params["blockId"]
        )
        return remote_path

    def upload_working_file(self, working_root, file_path):
        uploaded_files = self.upload_working_dir(working_root, file_path)
        return uploaded_files[0] if uploaded_files else None

    def upload_working_dir(self, working_root, local_path):
        if urlparse(local_path).scheme in ["s3", "s3n"]:
            return [local_path]
        remote_path = working_root
        return self.s3_upload(local_path, remote_path, recursive=True)

    def clean_working_dir(self, dir_path):
        self.s3_delete_file(dir_path)

    def prepare_working_file(self, working_root, file_path):
        return self.upload_working_file(working_root, file_path)

    def execute_jar(self, job_name, jar_path, jar_args=None, main_class="", *args, **kwargs):
        if not jar_args:
            jar_args = []
        print "EMR Jar job: '%s'" % jar_path
        print "           args = '%s'" % jar_args
        step_ids = self.emr_execute_jar(job_name=job_name,
                                        s3_jar_path=jar_path,
                                        jar_args=jar_args,
                                        main_class=main_class)
        self.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        self._wait_steps(step_ids, **kwargs)
        return 0

    def execute_hive(self, job_name, hive_script, *args, **kwargs):
        step_ids = self.emr_execute_hive(job_name, hive_script)
        self.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        self._wait_steps(step_ids, **kwargs)
        return step_ids

    def execute_pig(self, job_name, pig_script, *args, **kwargs):
        step_ids = self.emr_execute_pig(job_name, pig_script)
        self.emr_describe_steps(step_ids)

        print("Waiting jobflow steps...")
        self._wait_steps(step_ids, **kwargs)
        return step_ids


class QuboleMixin(BaseCluster):

    def prepare(self, hadoop_type, **cluster_kws):

        if hadoop_type not in ["QUBOLE"]:
            raise Exception("QuboleMixin :: Can NOT prepare '%s'" % hadoop_type)

        cluster = self.qubole_describe_cluster()
        cluster_ec2_settings = cluster['ec2_settings']
        self.aws_region = cluster_ec2_settings['aws_region']
        self.aws_key = cluster_ec2_settings['compute_access_key']
        self.aws_secret = cluster_ec2_settings['compute_secret_key']

    def get_working_root(self, cluster_params, global_params):
        remote_path = "s3n://{working_bucket}/zetjob/{username}/job{job_id}/blk{blk_id}/".format(
            working_bucket=cluster_params["S3_BUCKET"],
            username=global_params["userName"],
            job_id=global_params["jobId"],
            blk_id=global_params["blockId"]
        )
        return remote_path

    def upload_working_file(self, working_root, file_path):
        uploaded_files = self.upload_working_dir(working_root, file_path)
        return uploaded_files[0] if uploaded_files else None

    def upload_working_dir(self, working_root, local_path):
        if urlparse(local_path).scheme in ["s3", "s3n"]:
            return [local_path]
        remote_path = working_root
        return self.s3_upload(local_path, remote_path, recursive=True)

    def clean_working_dir(self, dir_path):
        self.s3_delete_file(dir_path)

    def prepare_working_file(self, working_root, file_path):
        return self.upload_working_file(working_root, file_path)

    def execute_jar(self, job_name, jar_path, jar_args, main_class, *args, **kwargs):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qubole_cluster_tags = self._kwargs.get("qubole_tags", None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)
        main_class = main_class if main_class else ""

        qbl_kwargs = {
            "name": job_name,
            "label": qubole_cluster_label,
            "tags": qubole_cluster_tags if qubole_cluster_tags else None,
            "sub_command": "jar",
            "sub_command_args": "%s %s %s" % (jar_path, main_class, " ".join(jar_args))
        }
        qubole_cmd = qds_sdk.commands.HadoopCommand.create(**qbl_kwargs)
        if not self._qubole_wait_command(qubole_cmd.id):
            raise Exception("QuboleCluster: failed to execute hadoop jar command, id=%s" % qubole_cmd.id)

    def execute_stream_jar(self, job_name, jar_file, jar_args, *args, **kwargs):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qubole_cluster_tags = self._kwargs.get("qubole_tags", None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)

        qbl_kwargs = {
            "name": job_name,
            "label": qubole_cluster_label,
            "tags": qubole_cluster_tags if qubole_cluster_tags else None,
            "sub_command": "streaming",
            "sub_command_args": "%s %s" % (jar_file, " ".join(jar_args))
        }
        qubole_cmd = qds_sdk.commands.HadoopCommand.create(**qbl_kwargs)

        if not self._qubole_wait_command(qubole_cmd.id):
            raise Exception("QuboleCluster: failed to execute hadoop streaming command, id=%s" % qubole_cmd.id)

    def execute_hive(self, job_name, hive_script, *args, **kwargs):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qubole_cluster_tags = self._kwargs.get("qubole_tags", None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)

        qbl_kwargs = {
            "name": job_name,
            "label": qubole_cluster_label,
            "script_location": hive_script,
            "tags": qubole_cluster_tags if qubole_cluster_tags else None
        }
        qubole_cmd = qds_sdk.commands.HiveCommand.create(**qbl_kwargs)

        if not self._qubole_wait_command(qubole_cmd.id):
            raise Exception("QuboleCluster: failed to execute hive command, id=%s" % qubole_cmd.id)

    def execute_pig(self, job_name, pig_script, *args, **kwargs):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qubole_cluster_tags = self._kwargs.get("qubole_tags", None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)

        qbl_kwargs = {
            "name": job_name,
            "label": qubole_cluster_label,
            "script_location": pig_script,
            "tags": qubole_cluster_tags if qubole_cluster_tags else None
        }
        qubole_cmd = qds_sdk.commands.PigCommand.create(**qbl_kwargs)

        if not self._qubole_wait_command(qubole_cmd.id):
            raise Exception("QuboleCluster: failed to execute pig command, id=%s" % qubole_cmd.id)

    def _qubole_wait_command(self, cmd_id):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)
        while True:
            time.sleep(10)
            ret = qds_sdk.commands.Command.find(cmd_id)
            print "Wait Qubole job: cluster_label='%s' step_id='%s' state='%s'" % \
                  (qubole_cluster_label, cmd_id, ret.status)
            if qds_sdk.commands.Command.is_done(ret.status):
                return qds_sdk.commands.Command.is_success(ret.status)

    def qubole_describe_cluster(self):
        qubole_api_token = self._kwargs['qubole_api_token']
        qubole_cluster_label = self._kwargs.get('qubole_cluster_name', None)
        qds_sdk.qubole.Qubole.configure(api_token=qubole_api_token)

        c = qds_sdk.cluster.Cluster.find(id=qubole_cluster_label)
        return c.attributes['cluster']


class EhcMixin(BaseCluster):

    def prepare(self, hadoop_type, **cluster_kws):

        if hadoop_type not in ["EHC"]:
            raise Exception("EhcMixin :: Can NOT prepare '%s'" % hadoop_type)

        self.ehc_client = self.create_ehc_client()
        user_info = self.ehc_client.get_user()
        cluster_ec2_settings = user_info['rawdata']
        self.aws_key = cluster_ec2_settings['s3AccessKeyId']
        self.aws_secret = cluster_ec2_settings['s3SecretAccessKey']

    def create_ehc_client(self):
        return EhcClient(ehc_id=self._kwargs['ehc_id'],
                         ehc_token=self._kwargs['ehc_token'])

    def get_working_root(self, cluster_params, global_params):
        remote_path = "s3n://{working_bucket}/zetjob/{username}/job{job_id}/blk{blk_id}/".format(
            working_bucket=cluster_params["S3_BUCKET"],
            username=global_params["userName"],
            job_id=global_params["jobId"],
            blk_id=global_params["blockId"]
        )
        return remote_path

    def upload_working_file(self, working_root, file_path):
        uploaded_files = self.upload_working_dir(working_root, file_path)
        return uploaded_files[0] if uploaded_files else None

    def upload_working_dir(self, working_root, local_path):
        if urlparse(local_path).scheme in ["s3", "s3n"]:
            return [local_path]
        remote_path = working_root
        return self.s3_upload(local_path, remote_path, recursive=True)

    def clean_working_dir(self, dir_path):
        self.s3_delete_file(dir_path)

    def prepare_working_file(self, working_root, file_path):
        return self.upload_working_file(working_root, file_path)

    def execute_jar(self, job_name, jar_path, jar_args, main_class, *args, **kwargs):
        pass

    def execute_stream_jar(self, job_name, jar_file, jar_args, *args, **kwargs):
        pass

    def execute_hive(self, job_name, hive_script, *args, **kwargs):
        pass

    def execute_pig(self, job_name, pig_script, *args, **kwargs):
        pass

    def hadoop_cmd(self, args, shell=False, verbose=True, extra_env=None):
        if verbose:
            print("Execute External Command : '%s'" % args)

        print " ".join(args)
        r = self.ehc_client.create_shell_command(" ".join(args))
        cmd_id = r["command_setting"][0]["id"]
        job_status = self.ehc_client.wait_command(cmd_id)
        if job_status != "succeeded":
            r = self.ehc_client.get_command_results(cmd_id)
            if verbose:
                print "--------------- EHC Logs ---------------"
                print r['results']
                print "----------------------------------------"
        ret = 0 if job_status == "succeeded" else -1

        if verbose:
            print("Exit with exit code = %d" % ret)
        return ret


class EmrCluster(EmrBaseCluster, EmrMixin):
    pass


class QuboleCluster(EmrBaseCluster, QuboleMixin):
    pass


class EhcCluster(EmrBaseCluster, EhcMixin):
    pass


class GenericHadoopCluster(BaseCluster):
    """A Generic Hadoop Cluster."""

    def __init__(self, **cluster_kws):
        super(GenericHadoopCluster, self).__init__()
        self.hs2_host = cluster_kws.get("hive_server2_host", None)
        self.hs2_port = cluster_kws.get("hive_server2_port", None)
        self.hadoop_conf_dir = self._get_default_hadoop_conf_dir(**cluster_kws)
        self.env_vars = {}

    def _get_default_hadoop_conf_dir(self, **cluster_kws):
        return cluster_kws.get("default_hadoop_conf_dir", "/etc/hadoop/conf/")

    def prepare(self, hadoop_type, **cluster_kws):
        if hadoop_type not in ["CDH4", "CDH5"]:
            raise Exception("GenericHadoopCluster :: Can NOT prepare '%s'" % hadoop_type)

        if "hadoop_conf" in cluster_kws:
            tmp_hadoop_conf_dir = prepare_hadoop_conf(cluster_kws["hadoop_conf"])
            if not tmp_hadoop_conf_dir:
                default_hadoop_conf_dir = self._get_default_hadoop_conf_dir(**cluster_kws)
                print "Use default hadoop_conf_dir '%s'" % default_hadoop_conf_dir
                self.hadoop_conf_dir = default_hadoop_conf_dir
            else:
                self.hadoop_conf_dir = tmp_hadoop_conf_dir

        env = os.environ.copy()
        cluster_def = cluster_kws.get("cluster_def_path", "/opt/cdh/env.json")
        self.env_vars = \
            preprocess_cluster_envs(base_envs=env,
                                    hadoop_type=hadoop_type,
                                    cluster_def=cluster_def)
        self.env_vars["HADOOP_CONF_DIR"] = self.hadoop_conf_dir

    def get_working_root(self, cluster_params, global_params):
        remote_path = "{hdfs_root}/zetjob/{username}/job{job_id}/blk{blk_id}/".format(
            hdfs_root=cluster_params["hdfs_root"],
            username=global_params["userName"],
            job_id=global_params["jobId"],
            blk_id=global_params["blockId"]
        )
        return remote_path

    def hadoop_cmd(self, args, shell=False, verbose=True, extra_env=None):
        if verbose:
            print("Execute External Command : '%s'" % args)
        env = os.environ.copy()
        env.update(self.env_vars)
        if extra_env:
            env.update(extra_env)

        ret = subprocess.call(args, shell=shell, env=env)
        if verbose:
            print("Exit with exit code = %d" % ret)
        return ret

    def upload_working_file(self, working_root, file_path):
        if self.hadoop_cmd(["hadoop", "fs", "-test", "-d", working_root]):
            self.hadoop_cmd(["hadoop", "fs", "-mkdir", "-p", working_root])

        if not self.hadoop_cmd(["hadoop", "fs", "-copyFromLocal", file_path, working_root]):
            return s3join(working_root, file_path)
        else:
            return None

    def upload_working_dir(self, working_root, local_path):
        if self.hadoop_cmd(["hadoop", "fs", "-test", "-d", working_root]):
            self.hadoop_cmd(["hadoop", "fs", "-mkdir", "-p", working_root])

        if not self.hadoop_cmd(["hadoop", "fs", "-copyFromLocal", local_path, working_root]):
            return s3join(working_root, local_path)
        else:
            return None

    def clean_working_dir(self, dir_path):
        return self.hadoop_cmd(["hadoop", "fs", "-rm", "-r", "-f", dir_path])

    def prepare_working_file(self, working_root, file_path):
        return file_path

    def execute_jar(self, job_name, jar_path, jar_args, main_class, *args, **kwargs):
        this_command = ["hadoop", "jar", jar_path]
        this_command.extend(jar_args)
        return self.hadoop_cmd(this_command)

    def execute_stream_jar(self, jar_file, jar_args, *args, **kwargs):
        pass

    def execute_hive(self, job_name, hive_script, **kwargs):
        if (not self.hs2_host) or (not self.hs2_port):
            raise Exception("Either 'hive_server2_host' or 'hive_server2_port' is not defined.")

        hive_db = kwargs.get("db", "default")
        hive_jdbc = "jdbc:hive2://%s:%s/%s" % (self.hs2_host, self.hs2_port, hive_db)
        return self.hadoop_cmd(["beeline", "-u", hive_jdbc,
                                "-n", "hive", "-p", "tiger",
                                "-d", "org.apache.hive.jdbc.HiveDriver",
                                "-f", hive_script, "--verbose=true"])

    def execute_pig(self, job_name, pig_script, *args, **kwargs):
        return self.hadoop_cmd(["pig", "-x", "mapreduce", pig_script])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

with open('./requirements.txt') as reqs_txt:
    requirements = [line for line in reqs_txt]

exec(open('datacanvas/version.py').read())

with open('./test-requirements.txt') as test_reqs_txt:
    test_requirements = [line for line in test_reqs_txt]

setup(
    name="pyDataCanvas",
    version=version,
    description="Runtime Support for DataCanvas.",
    packages=['datacanvas'],
    author="Xiaolin Zhang",
    author_email="leoncamel@gmail.com",
    install_requires=requirements,
    tests_require=test_requirements,
    zip_safe=False,
    test_suite='tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
    ],
)




import boto
from moto import mock_s3, mock_emr
from datacanvas.clusters import EmrCluster


def test_hello_world():
    pass

@mock_s3
@mock_emr
def test_emr_cluster():
    s3_conn = boto.connect_s3()
    # We need to create the bucket since this is all in Moto's 'virtual' AWS account
    s3_conn.create_bucket('mybucket')

    # EMR
    emr_conn = boto.connect_emr()
    cluster_id = emr_conn.run_jobflow(name="")
    print cluster_id

    # emr_cluster = EmrCluster(aws_region = "ap-southeast-1",
    #                          aws_key = "aws_key",
    #                          aws_secret = "aws_sec",
    #                          jobflow_id = cluster_id)

    emr_cluster = EmrCluster(aws_region = "ap-southeast-1",
                             aws_key = "aws_key",
                             aws_secret = "aws_sec",
                             jobflow_id = cluster_id)

    print "emr_conn...."
    print emr_cluster.emr_conn
    # print emr_cluster.s3_list_files("s3://mybucket/")
    # emr_cluster.execute_jar("job_name", "s3://mybucket/kaka.jar")




import re
import json
from collections import namedtuple


def _create_property(field_name, docstring, read_only=False, from_meta=False):
    """Helper for creating properties to IO objects to files.
    """

    def getter(self):
        if from_meta:
            if 'Meta' not in self:
                return None
            return self['Meta'].get(field_name, None)
        return self.get(field_name, None)

    def setter(self, value):
        if from_meta:
            if 'Meta' not in self:
                self['Meta'] = {}
            self['Meta'][field_name] = value
        else:
            self[field_name] = value

    def deleter(self):
        if from_meta:
            del self['Meta'][field_name]
        else:
            del self[field_name]

    if read_only:
        docstring = docstring + "\n\nThis attribute is read-only."

    if not read_only:
        return property(getter, setter, deleter, doc=docstring)
    return property(getter, doc=docstring)


class BaseIO(dict):
    Type = _create_property("Type", "", read_only=False, from_meta=False)

    def _init_attrs(self, attrs):
        for attr_name in attrs.keys():
            if hasattr(self.__class__, attr_name) and isinstance(getattr(DS_Hive, attr_name), property):
                setattr(self, attr_name, attrs.pop(attr_name))

    def __init__(self, seq=None, **attrs):
        self._init_attrs(attrs)
        if seq:
            super(BaseIO, self).__init__(seq)
        else:
            super(BaseIO, self).__init__(**attrs)


class DS_File(BaseIO):
    URL = _create_property("URL", "", read_only=False, from_meta=False)

    def __init__(self, seq=None, **attrs):
        if 'Type' not in attrs:
            attrs["Type"] = 'datasource.file'
        if seq:
            super(DS_File, self).__init__(seq)
        else:
            super(DS_File, self).__init__(**attrs)


class DS_S3(BaseIO):
    URL = _create_property("URL", "", read_only=False, from_meta=False)
    aws_key = _create_property("key", "", read_only=False, from_meta=True)
    aws_security = _create_property("token", "", read_only=False, from_meta=True)

    def __init__(self, seq=None, **attrs):
        if 'Type' not in attrs:
            attrs["Type"] = 'datasource.s3'
        if seq:
            super(DS_S3, self).__init__(seq)
        else:
            super(DS_S3, self).__init__(**attrs)


class DS_HDFS(BaseIO):
    URL = _create_property("URL", "", read_only=False, from_meta=False)
    port = _create_property("port", "", read_only=False, from_meta=True)

    def __init__(self, seq=None, **attrs):
        if 'Type' not in attrs:
            attrs["Type"] = 'datasource.hdfs'
        if seq:
            super(DS_HDFS, self).__init__(seq)
        else:
            super(DS_HDFS, self).__init__(**attrs)


class DS_Hive(BaseIO):
    URL = _create_property("URL", "", read_only=False, from_meta=False)
    meta_server = _create_property("hive_server2_host", "", read_only=False, from_meta=True)
    meta_port = _create_property("hive_server2_port", "", read_only=False, from_meta=True)

    def __init__(self, seq=None, **attrs):
        if 'Type' not in attrs:
            attrs["Type"] = 'datasource.hive'
        if seq:
            super(DS_Hive, self).__init__(seq)
        else:
            super(DS_Hive, self).__init__(**attrs)


class DS_Database(BaseIO):
    URL = _create_property("URL", "", read_only=False, from_meta=False)
    meta_type = _create_property("type", "", read_only=False, from_meta=True)
    meta_host = _create_property("host", "", read_only=False, from_meta=True)
    meta_port = _create_property("port", "", read_only=False, from_meta=True)
    meta_user = _create_property("user", "", read_only=False, from_meta=True)
    meta_password = _create_property("password", "", read_only=False, from_meta=True)
    meta_database = _create_property("database", "", read_only=False, from_meta=True)

    def __init__(self, seq=None, **attrs):
        if 'Type' not in attrs:
            attrs["Type"] = 'datasource.db'
        if seq:
            super(DS_Database, self).__init__(seq)
        else:
            super(DS_Database, self).__init__(**attrs)


_handler_callbacks = {}

TypeHandler = namedtuple("TypeHandler", ["handler", "type_pattern", "is_regex"])


def register_handler(type_name, type_pattern, handler, is_regex=False):
    if not type_pattern:
        type_pattern = type_name
    _handler_callbacks[type_name] = TypeHandler(handler=handler, type_pattern=type_pattern, is_regex=is_regex)


def specific_types():
    return dict((k, v) for k, v in _handler_callbacks.items() if not v.is_regex)


def generic_types():
    return dict((k, v) for k, v in _handler_callbacks.items() if v.is_regex)


def all_types():
    return _handler_callbacks


def is_subtype_of(type_name, target_type):
    gts = all_types()
    if type_name not in gts:
        return False
    type_handler = gts[type_name]
    if type_handler.is_regex:
        return re.match(type_handler.type_pattern, target_type)
    else:
        return type_handler.type_pattern == target_type


def is_type_of(type_name, target_type):
    return is_subtype_of(type_name, target_type) or is_subtype_of(type_name + ".*", target_type)


register_handler("LocalFile", None, DS_File)
register_handler("Http", None, DS_File)
register_handler("Ftp", None, DS_File)
register_handler("AWS_S3", None, DS_S3)
register_handler("HDFS", None, DS_HDFS)
register_handler("Hive", None, DS_Hive)
register_handler("DB", None, DS_Database)

register_handler("datasource.file", None, DS_File)
register_handler("datasource.s3", None, DS_S3)
register_handler("datasource.hdfs", None, DS_HDFS)
register_handler("datasource.hive", None, DS_Hive)
register_handler("datasource.db", None, DS_Database)
register_handler("datasource", None, BaseIO)

register_handler("s3", None, DS_S3)
register_handler("s3.*", "s3[\.\w+]+", DS_S3, is_regex=True)
register_handler("hdfs", None, DS_HDFS)
register_handler("hdfs.*", "hdfs[\.\w+]+", DS_HDFS, is_regex=True)
register_handler("hive", None, DS_Hive)
register_handler("hive.*", "hive[\.\w+]+", DS_Hive, is_regex=True)


def from_json(json_object):
    if 'Type' in json_object:
        jo_type = json_object['Type']

        stypes = specific_types()
        gtypes = generic_types()

        # Specific Type
        if jo_type in stypes:
            return stypes[jo_type].handler(json_object)
        else:
            # Generic Type
            for gt_name, gt_obj in gtypes.items():
                if re.match(gt_obj.type_pattern, jo_type):
                    return gt_obj.handler(json_object)
            else:
                return BaseIO(json_object)
    return json_object


def load_io_obj(filename):
    with open(filename, 'r') as f:
        return json.load(f, object_hook=from_json)
